(** This file is not necessary in the VPL library.
    
It only formalize weakest-liberal-preconditions and strongest-postconditions as
predicate transformers.
*)


Require Import PredTrans.
Require Import DemoPLVerifier.


Import ZCond.
Import Term.

Local Open Scope MPP.
Local Open Scope impure.
Local Open Scope statement.
Local Notation mem := (Mem.t ZNum.t).

(** * Weakest Liberal Precondition *)

Fixpoint wlp (s: statement): mem -> MPP mem :=
 match s with
 | (Assume c) => °-| (fun m => sat c m)
 | (Assert msg c) => °|- (fun m => sat c m)
 | (x ::= t) => Update (fun m => Mem.assign x (Term.eval t m) m)
 | (s1 -; s2) => (wlp s1) °; (wlp s2)
 | (If c Then s1 Else s2 Fi) => (°-| (sat c) °; (wlp s1)) °\/  (°-| (fun m => ~sat c m) °; (wlp s2))
 | (While c Do s Inv _ Done) => 
    UMeet
     (fun (inv: mem->Prop) => 
         °|- inv °;
         °|- (fun _ => forall m', sat c m' -> inv m' -> wlp s m' inv) °;
         (UJoin (fun m => Update (fun _ => m) °;
                         °-| (fun m => ~sat c m) °;
                         °-| inv)) )  
 end.
Local Hint Resolve apply_monotone.

Lemma wlp_correctness s m m': (sem s m m') -> forall post, wlp s m post -> wte m' post False.
Proof.
  induction 1; simpl in * |- *; intuition; try ((eapply IHsem || apply IHsem2; eauto; eapply IHsem1); eauto; simpl_ex; eauto 20).
  (* while_false *)
  - simpl_ex. 
  (* while_true_ko *)
  - simpl_ex.
Qed.

Local Hint Resolve sem_assume sem_assert_ok sem_assert_ko sem_assign sem_seq_ok sem_seq_ko 
  sem_ifs_true sem_ifs_false sem_while_false sem_while_true_ok sem_while_true_ko.

Ltac wlp_completude_simplify H := 
  let m' := fresh "m'" in
  evar (m': option mem); lapply (H m'); unfold m'; eauto; simpl; tauto.

Lemma wlp_completude s: forall m (post:mem -> Prop), (forall m', sem s m m' -> wte m' post False) -> wlp s m post.
Proof.
  induction s; simpl in * |- *;
  try (intuition; wlp_completude_simplify H).
  (* assert *)
  - intros; destruct (sat_dec c m); wlp_completude_simplify H.
  (* seq *)
  - intros; eapply IHs1. 
  intros m' H0; destruct m'; simpl.
    + eapply IHs2. intros; wlp_completude_simplify H.
    + wlp_completude_simplify H.
  (* while *)
  - intros; constructor 1 with (x:=fun m => forall m', sem (While c Do s Inv inv Done) m m' -> wte m' post False).
    intuition.
    + eapply IHs. intros m'0 H2. destruct m'0; simpl.
      * intros; wlp_completude_simplify H1.
      * wlp_completude_simplify H1.
    + wlp_completude_simplify H1.
Qed.

Theorem wlp_definition s post m: wlp s m post <-> ((is_ok s m) /\ (forall m', sem s m (Some m') -> post m')).
Proof.
  unfold is_ok; intuition; subst.
  - exact (wlp_correctness _ _ _ H0 _ H).
  - exact (wlp_correctness _ _ _ H0 _ H).
  - eapply wlp_completude.
    intros m' H; destruct m'; simpl; eauto.
Qed.  

Lemma is_ok_wlp s m: wlp s m (fun _ => True) <-> (is_ok s m).
Proof.
  rewrite wlp_definition.
  intuition.
Qed.

(** *  Strongest Postcondition *)

Fixpoint sp (s: statement): mem -> MPP mem :=
 match s with
 | (Assume c) => °|- (fun m => sat c m) 
 | (Assert msg c) => °-| (fun m => sat c m) 
 | (x ::= t) => UMeet (fun m0 => °|- (fun m => m=Mem.assign x (Term.eval t m0) m0) °; Update (fun _ => m0))
 | (s1 -; s2) => (sp s2) °; (sp s1)
 | (If c Then s1 Else s2 Fi) => ((sp s1) °; °|- (sat c)) °/\  ((sp s2) °; °|- (fun m => ~sat c m))
 | (While c Do s Inv _ Done) => 
    UJoin
     (fun (inv: mem->Prop) => 
         Fguard (fun _ => inv)
         (°-| (fun _ => forall m', sp s m' (fun m => sat c m /\ inv m)  -> inv m') °;
          °|- (fun m => ~sat c m)))
 end.

Lemma sp_correctness s m m': (sem s m m') -> forall (pre:mem -> Prop), pre m -> wte m' (fun m => sp s m pre) True.
Proof.
  induction 1; simpl in * |- *; try (intuition; eauto; xsimplify ltac:(intuition eauto); fail).
  (* while_true_ok *)
  intros pre H2. xsimplify ltac:(intuition eauto).
  + eapply (IHsem2 a0); eauto.
  + eapply (IHsem2 a0); eauto.
Qed.

Inductive isReachable c s: mem -> mem -> Prop  :=
  | isReachable_start m: isReachable c s m m
  | isReachable_next m0 m1 m2:  
    sat c m0
    -> sem s m0 (Some m1)

    -> isReachable c s m1 m2
       -> isReachable c s m0 m2.
Local Hint Resolve isReachable_start isReachable_next.

Lemma isReachable_While c s inv m0 m1: isReachable c s m0 m1 -> ~sat c m1 -> sem (While c Do s Inv inv Done) m0 (Some m1).
Proof.
  induction 1; simpl; eauto.
Qed.
Local Hint Resolve isReachable_While.

Lemma isReachable_trans c s m0 m1: isReachable c s m0 m1 -> forall m2, isReachable c s m1 m2 -> isReachable c s m0 m2.
Proof.
  induction 1; simpl; eauto.
Qed.
Local Hint Resolve isReachable_trans.


(** TODO: the proof is more complex...

Lemma sem_Some_Node_exclude s m m': sem s m m' -> m' <> None -> is_ok s m.
Proof.
  intros H. cut (wte m' (fun _ => wlp s m (fun _ => True)) True).
  destruct m'; simpl; rewrite <- is_ok_wlp; tauto.
  induction H; simpl in * |- *; auto; xsimplify ltac: (intuition; eauto).
  eapply apply_monotone. eauto.
  
  
Qed.

Lemma sp_completude s: forall (pre:mem -> Prop) m' m0, sp s m' pre -> 
  sem s m0 (Some m') -> exists m, pre m /\ sem s m (Some m').
Proof.
  induction s; simpl; try (intuition; subst; simpl_ex; fail).
  (* assert *)
  - intros. case (sat_dec c m'). 
    + intuition; simpl_ex.
    + 
  (* assign *)
  - intuition subst. simpl_ex. rewrite H0; eauto.
  (* seq *)
  - intros pre m' m0 H H0. destruct (IHs2 _ _ _ H) as [m1 H1].
    clear H; intuition.
    destruct (IHs1 _ _ H). intuition; simpl_ex.
  (* if *)
  - intros pre m' H; destruct H.
    destruct (IHs1 _ _ H). intuition; simpl_ex.
    destruct (IHs2 _ _ H). intuition; simpl_ex.
  (* while *)
  - intros pre m' H.
    destruct (H (fun m' => exists m, pre m /\ isReachable c s m m')).
    + intros; simpl_ex.
    + intros m0 H0. case (IHs _ _ H0); intuition.
      destruct H4.
      constructor 1 with (x:=x0).
      intuition eauto.
    + simpl_ex; eauto.      
Qed.

(** Galois connection with sp s (lower adjoint) and wlp s (upper adjoint)
*)

Local Hint Resolve wlp_completude sp_correctness. 

Lemma galois_connection s (pre post: mem -> Prop):
  (forall m', sp s m' pre -> post m') <-> (forall m, pre m -> wlp s m post).
Proof.
  intuition eauto.
  - destruct (sp_completude s _ _ H0); intuition.
    eapply wlp_correctness; eauto.
Qed.

Lemma galois_connection_1 s (post: mem -> Prop):
   forall m', sp s m' (fun m => wlp s m post) -> post m'.
Proof.
  rewrite galois_connection. auto.
Qed.
  
Lemma galois_connection_2 s (pre: mem -> Prop):
   forall m, pre m -> wlp s m (fun m' => sp s m' pre).
Proof.
  rewrite <- galois_connection. auto.
Qed.

Lemma remark_1 s1 s2 m:
   (wlp s1 m (fun m' => sp s2 m' (eq m))) -> (forall R, wlp s2 m R -> wlp s1 m R).
Proof.
  intros.
  apply wlp_completude.
  intros; eapply wlp_correctness. 2: eauto.
  destruct (sp_completude s2 (eq m) m').
  pattern m'; eapply wlp_correctness; eauto.
  intuition subst; auto.
Qed.

Lemma remark_2 s1 s2 m:
   (forall R, wlp s2 m R -> wlp s1 m R) -> (wlp s1 m (fun m' => sp s2 m' (eq m))).
Proof.
  intros.
  apply H.
  apply galois_connection_2. auto.
Qed.
**)
