(* Little test on affine extraction, polynomial normalization and polynomial equality

We compute affine extraction of "(1+x1+...+xN)^N"
  which is "1+N*(x1+...+xN)"

=> run with coqc.

Conclusion: affine extraction is quite good to decompose a polynomial into 
the sum of an affine expression and a polynomial with all monome of degre >= 2 !

*)

Require Import String.

Require Import Ring_polynom_AddOn.
Require Import ASTerm.
Import LinTerm.
Import ZAffTerm.

Module ZPeq := ZPomialEquality ZTerm ZTerm.
Import ZPeq.
Import M1.

Require Import List.
Definition PolExp (l:list ZTerm.t) (p:Pol): ZTerm.t 
 := Ring_polynom.Pphi (ZTerm.Cte 0) ZTerm.smartAdd ZTerm.smartMul ZTerm.Cte l p.

Import ZTerm.


Notation x1 := (Var 1%positive).
Notation x2 := (Var 2%positive).
Notation x3 := (Var 3%positive).
Notation x4 := (Var 4%positive).
Notation x5 := (Var 5%positive).
Notation x6 := (Var 6%positive).
Notation x7 := (Var 7%positive).
Notation x8 := (Var 8%positive).
Notation x9 := (Var 9%positive).
Notation x10 := (Var 10%positive).
Notation x11 := (Var 11%positive).
Notation x12 := (Var 12%positive).
Infix "+" := ZTerm.Add.
Notation "-" := ZTerm.Opp.
Infix "*" := ZTerm.Mul.

Notation "'°'" := Some.
Notation "'skipped'" := None.

Definition demo : bool := true. (* print polynomial X and Y below ! *)

Definition listvar := if demo then (x1::x2::x3::x4::x5::nil) else
                         (x1::x2::x3::x4::x5::x6::x7::x8::x9::x10::nil).


Definition idZ (x:Z): ZNum.t := x.
Coercion idZ: Z >-> ZNum.t.
Coercion Cte: ZNum.t >-> term.


(* PA = 1 + somme des var *)
Definition PA := Eval vm_compute in List.fold_left ZTerm.Add listvar 1.

(* PB = PA^(nombre de var) *)
Definition PB := Eval vm_compute in List.fold_left (fun p _ => smartMul PA p) listvar 1.

Definition affExtract te :=
  let (p,aft) := ZTerm.affineDecompose te in
  ZTerm.smartAdd (ZTerm.smartScalAdd (cte aft) (ZTerm.fromLin (lin aft)))
                  p.

Local Open Scope string.

Eval vm_compute in "PA:=(1+x1+...+xN):".
Eval vm_compute in PA.
Eval vm_compute in ("affine decomposition of PA").
Eval vm_compute in (affExtract PA).

Eval vm_compute in ("PB:=PA^N:").
Eval vm_compute in PB.

Eval vm_compute in ("X:=affine decomposition of PB").
Time Definition X := Eval vm_compute in (affExtract PB).

(* Normalisation of X is longer than Direct normalisation of PB 
Eval vm_compute in ("Normalisation of X").
Time Definition AuxX := Eval vm_compute in norm (M1.toPExpr X).
*)

Eval vm_compute in ("Direct Normalisation of PB").
Time Definition AuxY := Eval vm_compute in norm (M1.toPExpr PB). (* forme normale *)
Eval vm_compute in ("Y := Retranslation of the normalized polynom").
Time Definition Y := Eval vm_compute in PolExp listvar AuxY. (* retraduction sous forme polynomiale *)


(* time with printing the term *)
Eval vm_compute in ("print X").
Eval vm_compute in (if demo then Some X else None).
Eval vm_compute in ("print Y").
Eval vm_compute in (if demo then Some Y else None).

Eval vm_compute in ("eq Y PB").
Time Eval vm_compute in (pomial_eq Y PB).
Eval vm_compute in ("eq X PB").
Time Eval vm_compute in (pomial_eq X PB).
