type t = { id: int; f: Cert.frag_t; c: Cstr.t }

let get_id (x : t) = x.id
let get_f (x : t) = x.f
let get_c (x : t) = x.c

let mk id c = { id = id; f = Cert.mkF [id, Nb.u]; c = c }

let implies c1 c2 = Cstr.isIncl c1.c c2.c
let isEq c1 c2 = Cstr.isEq c1.c c2.c

let rename fromX toY c =
	let v = Cstr.get_v c.c in
	let v1 = Vec.set v fromX Nb.z in
	let v2 = Vec.set v1 toY (Vec.get v fromX) in
	assert (Nb.cmpz (Vec.get v toY) = 0);
	{c with c = {c.c with Cstr.v = v2}}

let joinSetup: bool -> Var.t -> Var.t option Rtree.t -> Var.t -> int -> t -> Var.t * Var.t option Rtree.t * t
= fun dup nxt relocTbl alpha idOffset c ->
	let (nxt1, vec1, relocTbl1) = Vec.shift nxt (Cstr.get_v c.c) relocTbl in
	let (vec2, alphaCoef, cst) =
		if dup then
			(Vec.add (Cstr.get_v c.c) (Vec.neg vec1), Cstr.get_c c.c, Cstr.get_c c.c)
		else
			(vec1, Nb.neg (Cstr.get_c c.c), Nb.z)
	in
	let c1 = {c.c with Cstr.v = Vec.set vec2 alpha alphaCoef; Cstr.c = cst} in
	(nxt1, relocTbl1, mk (idOffset + c.id) c1)

let initc c = { c with f = Cert.mkF [c.id, Nb.u] }
let extractc c = (c.id, Cert.Direct c.f)

let elimc id x c1 c2 =
	let (c, n1, n2) = Cstr.elim c1.c c2.c x in
	{ id = id; f = Cert.mergeF [c1.f, n1; c2.f, n2]; c = c }

let prExt: (Var.t -> string) -> t -> string
= fun varPr c -> Printf.sprintf "(%i): %s (frag: %s)" c.id (Cstr.pr varPr c.c) (Cert.prF c.f)

let pr: (Var.t -> string) -> t -> string
= fun varPr c -> Cstr.pr varPr c.c
