(** The information which is ultimately stored in the inclusion certificates is manipulated
explicitely during the calculations carried on polyhedra.
This module extends the representation of linear constraints to record this information on-the-fly. *)

type t = { id: int; f: Cert.frag_t; c: Cstr.t }

val get_id : t -> int
val get_f : t -> Cert.frag_t
val get_c : t -> Cstr.t

val mk: int -> Cstr.t -> t

val implies: t -> t -> bool
val isEq: t -> t -> bool
val rename: Var.t -> Var.t -> t -> t

val joinSetup: bool -> Var.t -> Var.t option Rtree.t -> Var.t -> int -> t -> Var.t * Var.t option Rtree.t * t

val initc: t -> t
val extractc: t -> int * Cert.cstr_t

val elimc: int -> Var.t -> t -> t -> t

val pr: (Var.t -> string) -> t -> string
val prExt: (Var.t -> string) -> t -> string
