(** This module is the interface between the VPL and the certified certificate checker.
It makes it easier to use the code extracted from the Coq development from Ocaml. *)

(** Convert a polyhedron from its VPL representation to an equivalent Coq representation. *)
val pToPedraQ: Pol.t -> PedraQ.BasicD.t

(** [polChkSyn p] checks that the constraints in [p.PedraQ.BasicD.cons] syntactically match these in [p.PedraQ.BasicD.ml], up to order. *)
val polChkSync: PedraQ.BasicD.t -> bool

(** [addC p c] performs the addition of [c] to [p] using the functions extracted from Coq. *)
val addC: Pol.t -> Cstr.t -> PedraQ.BasicD.t

(** Compute the intersection of two polyhedra using the
functions extracted from Coq. *)
val meet: Pol.t -> Pol.t -> PedraQ.BasicD.t

(** [isInclC p1 p2] checks whether the polyhedron [p1] is included in polyhedron [p2] by calling
the functions extracted from the certified Coq development.
If [isInclC] returns [true], then [p1] is included in [p2] and a proof of this fact was built. *)
val isInclC: Pol.t -> Pol.t -> bool

(** [joinC p1 p2] computes the convex hull of the polyhedra [p1] and [p2] using the certified Coq
abstract domain.
If [joinC] returns [None], then the result of the convex hull could not be computed while proving its correctness. *)
val joinC: Pol.t -> Pol.t -> PedraQ.BasicD.t

(** [getItvC p v] computes an interval of possible values for expression [v] from polyhedron [p]
using the functions extracted from Coq. *)
val getItvC: Pol.t -> Vec.t -> Pol.itvT

(** [project p x] computes the image of [p] by the projection of [x]
using the certified Coq abstract domain. *)
val project: Pol.t -> Var.t -> PedraQ.BasicD.t

(** [projectM p l] computes the image of [p] by the projection of all the
variables in [l] and checks the returned certificate. If certificate checking
fails, then the exception [Failure "checking error"] is raised. *)
val projectM: Pol.t -> Var.t list -> Pol.t
