type 'n t =
| Nil
| Sub of ('n t * 'n * 'n t)

type 'n rtree_t = 'n t

let rec get (z: 'n) (rt: 'n rtree_t) (v: Var.t) =
	match rt with
	| Nil -> z
	| Sub (l, n, r) ->
		match v with
		| Var.XH -> n
		| Var.XO t -> get z l t
		| Var.XI t -> get z r t

let set (z: 'n) (rt0: 'n rtree_t) (v0: Var.t) (n0: 'n) =
	let rec _set rt v: 'n rtree_t =
		match rt, v with
		| Nil, Var.XH -> Sub (Nil, n0, Nil)
		| Sub (l, _, r), Var.XH -> Sub (l, n0, r)
		| Nil, Var.XO t -> Sub (_set Nil t, z, Nil)
		| Sub (l, n, r), Var.XO t -> Sub (_set l t, n, r)
		| Nil, Var.XI t -> Sub (Nil, z, _set Nil t)
		| Sub (l, n, r), Var.XI t -> Sub (l, n, _set r t)
	in
		_set rt0 v0

let mk z l = List.fold_left (fun a (x, n) -> set z a x n) Nil l

let map (f: 'n -> 'm) (rt0: 'n rtree_t) =
	let rec _map rt =
		match rt with
		| Nil -> Nil
		| Sub (l, n, r) -> Sub (_map l, f n, _map r)
	in
		_map rt0

let rec fold (f: 'a -> 'n -> 'a) (a: 'a) (rt: 'n rtree_t): 'a =
	match rt with
	| Nil -> a
	| Sub (l, n, r) -> fold f (fold f (f a n) l) r

let rec find (f: 'a -> 'b option): 'a t -> (Var.t * 'b) option = function
	| Nil -> None
	| Sub (l, n, r) ->
		match f n with
		| Some b -> Some (Var.XH, b)
		| None ->

		match find f l with
		| Some (v, b) -> Some (Var.XO v, b)
		| None ->

		match find f r with
		| Some (v, b) -> Some (Var.XI v, b)
		| None -> None

(* XXX: GC? *)
let rec mskBuild1 : ('a -> bool) -> bool t -> 'a t -> bool t
= fun pred m -> function
	| Nil -> m
	| Sub (l, n, r) ->
		match m with
		| Nil -> Sub (mskBuild1 pred Nil l, pred n, mskBuild1 pred Nil r)
		| Sub (l', n', r') ->
			Sub (mskBuild1 pred l' l, n' || pred n, mskBuild1 pred r' r)

(* XXX: GC? *)
let mskBuild: ('a -> bool) -> 'a t list -> bool t
= fun pred l -> List.fold_left (mskBuild1 pred) Nil l

(* XXX: GC? *)
let rec pathsGet: bool t -> Var.t list
= function
	| Nil -> []
	| Sub (l, n, r) ->
		let lList = List.map (fun p -> Var.XO p) (pathsGet l) in
		let rList = List.map (fun p -> Var.XI p) (pathsGet r) in
		let sub = List.append lList rList in
		if n then Var.XH::sub else sub

(* XXX: GC? *)
let basisBuild: ('a -> bool) -> (Var.t -> string) -> 'a t list -> string t
= fun isNil pr l ->
	let nodes = List.map (fun p -> (p, pr p)) (pathsGet (mskBuild (fun n -> not (isNil n)) l)) in
	List.fold_left (fun tree (p, str) -> set "#inval" tree p str) Nil nodes

let rec toList: 'a t -> ('a * Var.t) list
= function
	| Nil -> []
	| Sub (l, n, r) ->
		let lList = List.map (fun (a, p) -> (a, Var.XO p)) (toList l) in
		let rList = List.map (fun (a, p) -> (a, Var.XI p)) (toList r) in
		(n, Var.XH)::(List.append lList rList)

let pr: string -> ('a -> string -> string) -> (Var.t -> string) -> 'a t -> string
= fun sep nodePr pathPr tree ->
	let nodeList = List.map (fun (a, p) -> nodePr a (pathPr p)) (toList tree) in
	String.concat sep (List.filter (fun s -> String.length s <> 0) nodeList)
