(** This module is the basis of the VPL.
The type [t] defined below is used in almost every module.
It is an association table of variable identifiers (see {! Var.t}) to some data. *)

type 'n t =
| Nil
| Sub of ('n t * 'n * 'n t)

val get: 'n -> 'n t -> Var.t -> 'n
val set: 'n -> 'n t -> Var.t -> 'n -> 'n t

val mk: 'n -> (Var.t * 'n) list -> 'n t

val map: ('n -> 'm) -> 'n t -> 'm t
val fold: ('a -> 'n -> 'a) -> 'a -> 'n t -> 'a

(** [find f rt] looks for a node [n] of [rt] for which [f] returns [Some b].
If such a node is found, the path to [n] in [rt] is returned along with the [b] returned by [f].
If [f] returns [None] for all the nodes of [rt], the function [find] returns [None].
The order in which [find] walks through [rt] is not specified. *)
val find: ('a -> 'b option) -> 'a t -> (Var.t * 'b) option

(** [mskBuild pred l] builds a mask tree where a node is true
if [pred] has returned [true] for at least one tree in[l] for that [node]. *)
val mskBuild: ('a -> bool) -> 'a t list -> bool t

(** [pathsGet tree] returns the list of the paths to node which value is [true]. *)
val pathsGet: bool t -> Var.t list

val toList: 'a t -> ('a * Var.t) list

(** [basisBuild isNil pr l] builds a [string t] where a path leads to its associated string (provided by [pr])
if there is at least a tree in [l] which associates a non-nil coefficient (determined by [isNil]) to this path. *)
val basisBuild: ('a -> bool) -> (Var.t -> string) -> 'a t list -> string t

val pr: string -> ('a -> string -> string) -> (Var.t -> string) -> 'a t -> string
