(** This module provides the necessary elements for the Coq development
to call the VPL to build proofs of its results. *)

val coqPosToZ: BinNums.positive -> Nb.Z.t
val zToCoqPos: Nb.Z.t -> BinNums.positive

val coqZToZ: BinNums.coq_Z -> Nb.Z.t
val zToCoqZ: Nb.Z.t -> BinNums.coq_Z

val nToNb: NumC.QNum.t -> Nb.t
val nToNumC: Nb.t -> NumC.QNum.t

val progVarToVar: ProgVar.PVar.t -> Var.t
val varToProgVar: Var.t -> ProgVar.PVar.t

val ltToVec: LinTerm.LinQ.t -> Vec.t
val vecToLt: Vec.t -> LinTerm.LinQ.t

val cToCstr: CstrC.Cstr.t -> Cstr.t
val cToCstrC: Cstr.t -> CstrC.Cstr.t
val cToConsC: Cons.t -> ConsSet.Cs.cT

val pToConsSet: Pol.t -> ConsSet.Cs.t

val cstrPr: CertC.Cert.cstrT -> string
val cToCertC: Cert.t -> CertC.Cert.t
val certCPr: CertC.Cert.t -> string
val joinEltPr: CertC.Cert.joinEltT -> string
val joinPr: CertC.Cert.joinT -> string

val iToItv: Itv.QItv.t -> Pol.itvT

(** Check whether a [ConsSet.Cs.t] and a [Pol.t] describe exactly the
same polyhedra, in such a way that further communication between the
code extracted from Coq and the Ocaml code has a chance to work. *)
val chkSync: ConsSet.Cs.t -> Pol.t -> bool

val isEmpty: Pol.t -> CertC.Cert.t option
val isIncl: Pol.t * Pol.t -> CertC.Cert.t option
val add: Pol.t * CstrC.Cstr.t -> Pol.t option * CertC.Cert.t
val addp: Pol.t * CstrC.Cstr.t -> Backend.addPrecT
val meet: Pol.t * Pol.t -> Pol.t option * CertC.Cert.meetT
val join: Pol.t * Pol.t -> Pol.t * CertC.Cert.joinT
val widen: Pol.t * Pol.t -> Pol.t * ConsSet.Cs.t
val getItv: Pol.t * LinTerm.LinQ.t -> Backend.itvT
val getLowerBound: Pol.t * LinTerm.LinQ.t -> Backend.bndT
val getUpperBound: Pol.t * LinTerm.LinQ.t -> Backend.bndT
val project: Pol.t * ProgVar.PVar.t -> Pol.t * CertC.Cert.t
val rename : (ProgVar.PVar.t * ProgVar.PVar.t) * Pol.t -> Pol.t
