type t = XH | XO of t | XI of t

let toInt: t -> int
= fun p0 ->
	let inc i =
		let msb = Pervasives.max_int - (Pervasives.max_int lsr 1) in
		if i = msb then (* overflow *)
			Pervasives.invalid_arg "Var.toInt"
		else
			i lsl 1
	in
	let rec fn i msb = function
		| XH -> msb + i
		| XO p -> fn i (inc msb) p
		| XI p -> fn (msb + i) (inc msb) p
	in
	fn 0 1 p0

let fromInt: int -> t
= fun i0 ->
	let rec _fromInt i =
		let iMasked = i land 1 in
		if i = iMasked then
			XH
		else
			let tail = _fromInt (i lsr 1) in
			if iMasked = 1 then
				XI tail
			else
				XO tail
	in
	if i0 > 0 then
		_fromInt i0
	else
		Pervasives.invalid_arg "Var.fromInt"

let pr': string -> t -> string
= fun s p -> s ^ (Pervasives.string_of_int (toInt p))

let pr : t -> string
= fun p -> pr' "v" p

let next (bp: t): t =
	let rec next_rec (bp': t): t * bool =
		match bp' with
		  XH -> (XH, true)
		| XI tail -> (
			match next_rec tail with
			  (tail', true) -> (XO tail', true)
			| (tail', false) -> (XI tail', false))
		| XO tail -> (
			match next_rec tail with
			  (tail', true) -> (XI tail', false)
			| (tail', false) -> (XO tail', false))
	in
	let (res, overflow) = next_rec bp in
	if overflow then XO res else res

let cmp: t -> t -> int
= fun x00 x01 ->
	let rec _cmp x0 x1 dec res =
		match x0, x1 with
		| XO tl0, XO tl1 -> _cmp tl0 tl1 dec res
		| XI tl0, XI tl1 -> _cmp tl0 tl1 dec res
		| XI tl0, XO tl1 -> _cmp tl0 tl1 true (if dec then res else 1)
		| XO tl0, XI tl1 -> _cmp tl0 tl1 true (if dec then res else -1)
		| XH, XO _ | XH, XI _ -> -1
		| XO _, XH  | XI _, XH -> 1
		| XH, XH -> res
	in
	_cmp x00 x01 false 0

let compare : t -> t -> int
= cmp

let horizon: t list -> t
= function
	| [] -> XH
	| _::_ as l ->
		let sortedVars = List.sort (fun x1 x2 -> - (cmp x1 x2)) l in
		next (List.hd sortedVars)
