open Otypes

	type prophecy = ASTerm.BasicZTerm.term list
	
	type t = Polynomial.PolyZ.t -> env -> mode -> Polynomial.VariablesInt.t MapMonomial.t -> (AnnotedVar.t list) MapMonomial.t -> Pattern.t -> (prophecy * Polynomial.VariablesInt.t MapMonomial.t * (AnnotedVar.t list) MapMonomial.t * Polynomial.PolyZ.t)

	let rec (default : t)
		= fun p env mode mapKeep mapNKeep pat ->
		match pat with
		| Pattern.Default p -> (match p with
			| [] -> ([], mapKeep, mapNKeep, [])
			| (m,c) :: tl -> let (pro', mapKeep', mapNKeep', poly') = default p env mode mapKeep mapNKeep (Default tl) in
			let var_to_keep = try MapMonomial.find m mapKeep'
				with Not_found _ -> List.nth (AnnotedVar.update_monomial m mapNKeep') 0 in
			let aVarl = try MapMonomial.find m mapNKeep'
				with Not_found _ -> [] in
			let aff = ASTerm.BasicZTerm.annotAFFINE (Term.of_monomial ([var_to_keep], c)) in
			let mon = ASTerm.BasicZTerm.smartAnnot ASTerm.TopLevelAnnot.INTERV
				(Term.of_monomialBasis (Polynomial.Misc.pop Polynomial.VariablesInt.equal m var_to_keep)) 
				|> fun t -> AnnotedVar.apply t aVarl in
			let value = ASTerm.BasicZTerm.smartMul mon aff in
			( value :: pro', mapKeep', mapNKeep', poly'))
		| _ -> Pervasives.failwith "Heuristic.default"

	let (varCte : t)
		= fun p env mode mapKeep mapNKeep pat ->
		match pat with 
		| Pattern.VarCte((m,c),v) -> let l = 
			try MapMonomial.find m mapNKeep
			with Not_found _ -> [] in
			([], mapKeep, MapMonomial.add m ((AnnotedVar.of_var v ASTerm.TopLevelAnnot.STATIC) :: l) mapNKeep, p)
		| _ -> Pervasives.failwith "Heuristic.varCte"

	let (multiplicity : t)
		= 	let rec(range : int -> int -> int list)
			= fun i j->
			if i >= j then [] else i :: (range (i+1) j)
		in fun p env mode mapKeep mapNKeep pat ->
		match pat with 
		| Pattern.Multiplicity(l,v) -> 
		let mapNKeep' = 
		List.fold_left
		(fun map ((m,c),i) -> let l' = try MapMonomial.find m map with Not_found _ -> [] in
		let l'' =  List.map (fun j -> AnnotedVar.Var v) (range 0 i) in
		MapMonomial.add m (l''@l') map)
		mapNKeep
		l
		in ([], mapKeep, mapNKeep', p)
		| _ -> Pervasives.failwith "Heuristic.multiplicity"
		
	let (unboundedVarmode : t)
		= fun p env mode mapKeep mapNKeep pat ->
		match pat with 
		| Pattern.UnboundedVarmode((m,c),v) -> let l = 
			try MapMonomial.find m mapNKeep
			with Not_found _ -> [] in
			([], mapKeep, MapMonomial.add m (AnnotedVar.Var v :: l) mapNKeep, p)
		| _ -> Pervasives.failwith "Heuristic.unboundedVarmode"
			
	let (greatestItv : t)
		= fun p env mode mapKeep mapNKeep pat ->
		match pat with
		| Pattern.GreatestItv((m,c),v) -> ([], MapMonomial.add m v mapKeep, mapNKeep, p)
		| _ -> Pervasives.failwith "Heuristic.greatestItv"

	let (unboundedVar : t)
		= fun p env mode mapKeep mapNKeep pat ->
		match pat with
		| Pattern.UnboundedVar((m,c),v) -> ([], MapMonomial.add m v mapKeep, mapNKeep, p)
		| _ -> Pervasives.failwith "Heuristic.unboundedVar"		

	let (linearMonomial : t)
		= fun p env mode mapKeep mapNKeep pat ->
		match pat with
		| Pattern.LinearMonomial ((m,c),v) -> ([], MapMonomial.add m v mapKeep, mapNKeep, p)
		| _ -> Pervasives.failwith "Heuristic.linearMonomial"

	let (centerZero : t)
		= fun p env mode mapKeep mapNKeep pat ->
		match pat with
		| Pattern.CenterZero(m,c) -> let vToKeep = MapMonomial.find m mapKeep in
		let aVarl = try MapMonomial.find m mapNKeep with Not_found _ -> [] in
		let l = Polynomial.Misc.pop Polynomial.VariablesInt.equal m vToKeep
		|> List.filter (fun v -> v |> Polynomial.VariablesInt.to_int > 0 && 
			(v |> Itv.of_var env |> fun itv -> Itv.is_bounded itv && not (Itv.contains_zero itv)) &&
			(not (AnnotedVar.mem v aVarl))) in
		let tlist = Term.center_zero_var (m,c) vToKeep l (List.map (fun v -> v |> Itv.of_var env |> Itv.get_translation_bound) l)
		|> List.map (fun t -> AnnotedVar.apply t aVarl) in
		(tlist, mapKeep, mapNKeep, Polynomial.PolyZ.remove_monomial p m)
		| _ -> Pervasives.failwith "Heuristic.centerZero"		

	let (translation : t)
		= fun p env mode mapKeep mapNKeep pat ->
		match pat with
		| Pattern.Translation(m,c) ->
		let aVarl = try MapMonomial.find m mapNKeep with Not_found _ -> [] in
		let vToKeep = MapMonomial.find m mapKeep in
		let (t, m') = Term.translate (m,c) vToKeep (vToKeep |> Itv.of_var env |> Itv.get_translation_bound) in
		let t' = AnnotedVar.apply t aVarl in
		([t'], mapKeep, mapNKeep, Polynomial.PolyZ.add (Polynomial.PolyZ.remove_monomial p m) [m'])
		| _ -> Pervasives.failwith "Heuristic.translation"
	
	let (screwed : t)
		= fun p env mode mapKeep mapNKeep pat ->
		match pat with
		| Pattern.Screwed -> default p env mode mapKeep mapNKeep (Pattern.Default p)
		| _ -> Pervasives.failwith "Heuristic.screwed"

	let (monomialCte : t)
		= fun p env mode mapKeep mapNKeep pat ->
		match pat with
		| Pattern.MonomialCte (m,c) -> 
			([Term.of_monomial (m,c)], mapKeep, mapNKeep, Polynomial.PolyZ.remove_monomial p m)
		| _ -> Pervasives.failwith "Heuristic.monomialCte"

	let (firstUnbounded : t)
		= fun p env mode mapKeep mapNKeep pat ->
		match pat with
		| Pattern.FirstUnbounded ((m,c),v) -> ([], MapMonomial.add m v mapKeep, mapNKeep, p)
		| _ -> Pervasives.failwith "Heuristic.firstUnbounded"
		
	let (nulScalar : t)
		= fun p env mode mapKeep mapNKeep pat ->
		match pat with
		| Pattern.NulScalar (m,c) -> 
			([], mapKeep, mapNKeep, Polynomial.PolyZ.remove_monomial p m)
		| _ -> Pervasives.failwith "Heuristic.nulScalar"
		
	let (of_pattern : t)
		= fun p env mode mapKeep mapNKeep pat ->
		Misc.trace "Heuristic used : ";
		match pat with
		| Pattern.NulScalar _ -> Misc.trace "NulScalar\n"; nulScalar p env mode mapKeep mapNKeep pat
		| Pattern.Multiplicity _ -> Misc.trace "Multiplicity\n"; multiplicity p env mode mapKeep mapNKeep pat
		| Pattern.MonomialCte _ -> Misc.trace "MonomialCte\n"; monomialCte p env mode mapKeep mapNKeep pat
		| Pattern.CenterZero _ -> Misc.trace "CenterZero\n";centerZero p env mode mapKeep mapNKeep pat
		| Pattern.Translation _ -> Misc.trace "Translation\n"; translation p env mode mapKeep mapNKeep pat
		| Pattern.LinearMonomial _ -> Misc.trace "LinearMonomial\n"; linearMonomial p env mode mapKeep mapNKeep pat
		| Pattern.Screwed _ -> Misc.trace "Screwed\n"; screwed p env mode mapKeep mapNKeep pat
		| Pattern.VarCte _ -> Misc.trace "VarCte\n"; varCte p env mode mapKeep mapNKeep pat
		| Pattern.GreatestItv _ -> Misc.trace "GreatestItv\n"; greatestItv p env mode mapKeep mapNKeep pat
		| Pattern.UnboundedVar _ -> Misc.trace "UnboundedVar\n"; unboundedVar p env mode mapKeep mapNKeep pat
		| Pattern.UnboundedVarmode _ -> Misc.trace "UnboundedVarmode\n"; unboundedVarmode p env mode mapKeep mapNKeep pat
		| Pattern.FirstUnbounded _ -> Misc.trace "FirstUnbounded\n"; firstUnbounded p env mode mapKeep mapNKeep pat
		| Pattern.Default _ -> Misc.trace "Default\n"; default p env mode mapKeep mapNKeep pat

