
type env = ProgVar.PVar.t -> ZNoneItv.ZNItv.t

type mode = Both | Up | Low

module MapMonomial : Map.S with type key = Polynomial.PolyZ.MonomialBasis.t

module Misc
	: sig
	
	val listToString : ('a -> string) -> 'a list -> string -> string

  	val map_to_string : 'a MapMonomial.t -> ('a -> string) -> string
	
	val trace : string -> unit
end

module Term
	: sig
	
	type t = ASTerm.BasicZTerm.term
	
	val zero : t
	
	val one : t
	
	val to_polynomial: t -> Polynomial.PolyZ.t
	
	val to_string : t -> string
	
	val of_cte : Polynomial.ScalarZ.t -> t
		
	val of_var : Polynomial.VariablesInt.t -> t
	
	val of_monomialBasis : Polynomial.PolyZ.MonomialBasis.t -> t

	val of_monomial : Polynomial.PolyZ.Monomial.t ->t
	
	val of_polynomial : Polynomial.PolyZ.t -> t
		
	val center_zero_var : Polynomial.PolyZ.Monomial.t -> Polynomial.VariablesInt.t -> Polynomial.VariablesInt.t list-> BinNums.coq_Z list -> ASTerm.BasicZTerm.term list
		
	val get_affine_part : t -> t
  	
	val get_interv_part : t -> t
	
	val translate : Polynomial.PolyZ.Monomial.t -> Polynomial.VariablesInt.t -> BinNums.coq_Z -> ASTerm.BasicZTerm.term * Polynomial.PolyZ.Monomial.t

end

(* les annotations autorisées sont Interv et Static*)
module AnnotedVar
	: sig
	
	type t = 
	Var of Polynomial.VariablesInt.t
	| AVar of ASTerm.TopLevelAnnot.topLevelAnnot * t
	
	val of_var : Polynomial.VariablesInt.t -> ASTerm.TopLevelAnnot.topLevelAnnot -> t
	
	val to_var : t -> Polynomial.VariablesInt.t
		
	val to_term : t -> ASTerm.BasicZTerm.term
		
	val to_string : t -> string
	
	(* utile pour prendre en compte les variables éliminées pour un monôme
	il faut prendre garde à ce que le pattern fournisse le monome original cependant *)
	val update_monomial : Polynomial.PolyZ.MonomialBasis.t -> (t list) MapMonomial.t -> Polynomial.PolyZ.MonomialBasis.t
	
	val apply : Term.t -> t list -> Term.t
	
	val mem : Polynomial.VariablesInt.t -> t list -> bool
	
  	val find : Polynomial.VariablesInt.t -> t list -> t
end

module Itv
	: sig
	
	type t = ZNoneItv.ZNItv.t
	
	val low : t -> ZNone.ZN.t
	
	val up : t -> ZNone.ZN.t
		
	val of_var : env -> Polynomial.PolyZ.var -> t
	
	val of_term : Term.t -> env -> t
	
	val to_string : t -> string
			
	val range : t -> int
	
	val is_bounded : t -> bool

	val is_fully_unbounded : t -> bool
	
	val greatest : Polynomial.PolyZ.MonomialBasis.t -> env -> Polynomial.PolyZ.var
	
	val get_center : t -> BinNums.coq_Z
	
	val contains_zero : t -> bool
	
	val get_translation_bound : t -> BinNums.coq_Z
end
