type 'a mayFailT
=
	| Ok of 'a
	| Failed of string

let process: Dat.t mayFailT -> string -> Dat.t mayFailT
= fun dat m ->
	match dat with
	| Failed _ as err -> err
	| Ok dat ->
		match Dep.gather m with
		| Dep.Failed err -> Failed err
		| Dep.Ok deps ->
			match Dat.depends m deps dat with
			| Dat.Failed err -> Failed err
			| Dat.Ok dat1 -> Ok dat1

let main: string array -> unit
= fun args ->
  match Array.to_list args |> List.tl with
  | [] -> Pervasives.failwith "need at least one file name"
  | _ :: _ as files ->
     match List.fold_left process (Ok Dat.nil) files with
     | Failed err -> Pervasives.prerr_endline err
     | Ok dat ->
	match Dat.order dat with
	| Dat.Failed err -> (
	  Pervasives.prerr_endline err;
	  Pervasives.exit 1
	)
	| Dat.Ok l -> Pervasives.print_string (String.concat " " l)

let _ = main Sys.argv
